#define _PHIDGET_NETWORKCODE
#include "../server.h"
#include "mos/mos_readdir.h"

static PhidgetServerHandle netconn;

static PhidgetReturnCode
setNetworkFlags(pconf_t *cfg) {

	if (pconf_getbool(cfg, 0, "phidget.network.resolveaddrs"))
		PhidgetNet_setProperty("resolveaddrs", "true");

	return (EPHIDGET_OK);
}

static PhidgetReturnCode
readDictionary(mosiop_t iop, const char *dir, const char *file) {
	PhidgetDictionaryHandle dict;
	PhidgetReturnCode res;
	const char *bname;
	const char *label;
	char val[1024];
	pconf_t *pc;
	int sn;
	int i;

	dict = NULL;

	res = pconf_parsepc(&pc, val, sizeof (val), "%s/%s", dir, file);
	if (res != EPHIDGET_OK)
		return (MOS_ERROR(iop, res, "failed to parse dictionary '%s': %s", file, val));

	/*
	 * If disabled cleanup and return.
	 */
	if (!pconf_getbool(pc, 0, "dictionary.enabled"))
		goto bad;

	label = pconf_getstr(pc, NULL, "dictionary.label");
	if (label == NULL) {
		res = MOS_ERROR(iop, EPHIDGET_INVALID, "missing 'dictionary.label'");
		goto bad;
	}

	sn = pconf_get32(pc, -1, "dictionary.sn");
	if (sn == -1) {
		res = MOS_ERROR(iop, EPHIDGET_INVALID, "missing 'dictionary.sn'");
		goto bad;
	}

	res = PhidgetDictionary_addDictionary(sn, label);
	if (res != EPHIDGET_OK) {
		nslogerr("failed to add dictionary %s/%d: %d", label, sn, res);
		goto bad;
	}

	/*
	 * If disabled cleanup and return.
	 */
	if (!pconf_exists(pc, "dictionary.config.key"))
		goto bad;

	res = PhidgetDictionary_create(&dict);
	if (res != EPHIDGET_OK) {
		nslogerr("failed to create dictionary handle for %s/%d", label, sn);
		goto bad;
	}

	res = Phidget_setDeviceSerialNumber((PhidgetHandle)dict, sn);
	if (res != EPHIDGET_OK) {
		nslogerr("failed to set device serial number on dictionary %s/%d", label, sn);
		goto bad;
	}
	res = Phidget_setDeviceLabel((PhidgetHandle)dict, label);
	if (res != EPHIDGET_OK) {
		nslogerr("failed to set device label on dictionary %s/%d", label, sn);
		goto bad;
	}

	res = Phidget_openWaitForAttachment((PhidgetHandle)dict, 2000);
	if (res != EPHIDGET_OK) {
		nslogerr("failed to open dictionary %s/%d: %d", label, sn, res);
		goto bad;
	}

	for (i = 0; ; i++) {
		bname = pconf_getentryname(pc, i, "dictionary.config.key");
		if (bname == NULL)
			break;

		if (!Phidget_validDictionaryKey(bname)) {
			nslogerr("invalid dictionary key: '%s': ignoring", bname);
			continue;
		}

		res = pconf_tostring(pc, val, sizeof (val), "dictionary.config.key.%s.value", bname);
		if (res != EPHIDGET_OK)
			continue;

		res = PhidgetDictionary_add(dict, bname, val);
		if (res != EPHIDGET_OK)
			nslogerr("failed to add key '%s' on dictionary %s/%d", bname, label, sn);
	}

	res = EPHIDGET_OK;
	Phidget_close((PhidgetHandle)dict);

bad:

	if (dict)
		PhidgetDictionary_delete(&dict);
	if (pc)
		pconf_release(&pc);

	return (res);
}

static PhidgetReturnCode
loadDictionaries(pconf_t *pc) {
	mos_dirinfo_t *di;
	const char *dir;
	int err;

	if (!pconf_getbool(pc, 1, "phidget.feature.dictionary.enabled"))
		return (EPHIDGET_OK);

	dir = pconf_getstr(pc, DICTCONFIGDIR, "phidget.feature.dictionary.directory");

	err = mos_opendir(MOS_IOP_IGNORE, dir, &di);
	if (err != 0) {
		nslogerr("failed to open directory '%s'", dir);
		return (err);
	}

	for (;;) {
		err = mos_readdir(MOS_IOP_IGNORE, di);
		if (err != 0) {
			nslogerr("failed to read directory '%s'", dir);
			goto bad;
		}

		if (di->errcode == MOSN_NOENT)
			break;

		if (di->flags & MOS_DIRINFO_ISDIR)
			continue;

		if (!mos_endswith(di->filename, DICTCONFIGEXT))
			continue;

		nsloginfo("loading dictionary %s", di->filename);
		err = readDictionary(MOS_IOP_IGNORE, dir, di->filename);
		if (err != EPHIDGET_OK)
			nslogerr("failed to read dictionary %s", di->filename);
	}

bad:

	mos_closedir(&di);
	return (err);
}

static void
blockv4Clients(pconf_t *cfg) {
	const char *addr;
	char json[64];
	int verbose;
	int i;

	if (!pconf_getbool(cfg, 1, "phidget.filter.clients.deny.enabled")) {
		nsloginfo("client deny filtering disabled");
		return;
	}

	verbose = pconf_getbool(cfg, 0, "phidget.filter.clients.deny.verbose");

	for (i = 0; ; i++) {
		addr = pconf_getstr(cfg, NULL, "phidget.filter.clients.deny.list.%d", i);
		if (addr == NULL)
			break;
		mos_snprintf(json, sizeof (json), "{\"family\":%d, \"addr\":\"%s\",\"verbose\":%d}",
		  AF_INET, addr, verbose);
		if (PhidgetNet_setProperty("blockclient", json) != EPHIDGET_OK)
			nslogerr("failed to add block for client: '%s'", addr);
	}
}

static void
allowv4Clients(pconf_t *cfg) {
	const char *addr;
	char json[64];
	int verbose;
	int i;

	if (!pconf_getbool(cfg, 1, "phidget.filter.clients.allow.enabled")) {
		nsloginfo("client allow filtering disabled");
		return;
	}

	verbose = pconf_getbool(cfg, 0, "phidget.filter.clients.allow.verbose");

	for (i = 0; ; i++) {
		addr = pconf_getstr(cfg, NULL, "phidget.filter.clients.allow.list.%d", i);
		if (addr == NULL)
			break;
		mos_snprintf(json, sizeof (json), "{\"family\":%d, \"addr\":\"%s\",\"verbose\":%d}",
		  AF_INET, addr, verbose);
		if (PhidgetNet_setProperty("allowclient", json) != EPHIDGET_OK)
			nslogerr("failed to add block for client: '%s'", addr);
	}
}

static void
filterClients(pconf_t *cfg) {

	if (pconf_getbool(cfg, 1, "phidget.filter.enabled") == 0)
		return;

	if (mos_strcmp(pconf_getstr(cfg, "allow", "phidget.filter.clients.default"), "allow") == 0)
		PhidgetNet_setProperty("allowclients", "%d", 1);
	else
		PhidgetNet_setProperty("allowclients", "%d", 0);
	blockv4Clients(cfg);
	allowv4Clients(cfg);
}

PhidgetReturnCode
startPhidgetServer(pconf_t *cfg) {
	PhidgetReturnCode res;
	const char *srvname;
	const char *passwd;
	char compname[64];
	int flags;

	const char *ipv4_addr;
	int ipv4_port;
	int keepalive;
	int allowdg;

	getComputerName(compname, sizeof (compname), "Phidget22Server");
	flags = 0;

	srvname = pconf_getstr(cfg, compname, "phidget.network.publish.name");
	passwd = pconf_getstr(cfg, "", "phidget.auth.password");
	ipv4_port = pconf_get32(cfg, -1, "phidget.network.ipv4.port");
	ipv4_addr = pconf_getstr(cfg, NULL, "phidget.network.ipv4.address");
	keepalive = pconf_get32(cfg, -1, "phidget.network.keepalive");
	allowdg = pconf_getbool(cfg, 1, "phidget.network.datagram.enabled");

	if (allowdg)
		PhidgetNet_setProperty("allowdatagram", "true");
	else
		PhidgetNet_setProperty("allowdatagram", "false");

	if (keepalive >= 0) {
		nsloginfo("Changed network keepalive to %d", keepalive);
		PhidgetNet_setProperty("keepalive", "%d", keepalive);
	}

	if (pconf_getbool(cfg, 0, "phidget.feature.stats.enabled"))
		PhidgetDictionary_enableStatsDictionary();

	if (pconf_getbool(cfg, 1, "phidget.feature.control.enabled"))
		PhidgetDictionary_enableControlDictionary();

	if (pconf_getbool(cfg, 1, "phidget.network.publish.enabled"))
		flags |= PHIDGET_NETWORK_PUBLISHMDNS;

	filterClients(cfg);
	setNetworkFlags(cfg);
	loadDictionaries(cfg);

	if (ipv4_port != -1) {
		res = PhidgetNet_startServer(flags, AF_INET, srvname, ipv4_addr, ipv4_port, passwd, &netconn);
		if (res != EPHIDGET_OK) {
			nsloginfo("PhidgetNet_startserver() failed on IPv4 port %d", ipv4_port);
			return (res);
		}
		nsloginfo("Started Phidget Server %s %s:%d", srvname, ipv4_addr ? ipv4_addr : "any", ipv4_port);
	}

	return (EPHIDGET_OK);
}

void
stopPhidgetServer() {

	PhidgetNet_stopServer(&netconn);
}