#include "webserver.h"
#include "sqlite3.h"
#include "mos/mos_fileio.h"

typedef PhidgetReturnCode (*dbcheck_t)(pconf_t *, pconf_t *, const char *);

struct arg {
	const char *name;
	const char *col;
	const char *cmp;
	int			type;
	dbcheck_t	check;
};

static PhidgetReturnCode validString(pconf_t *, pconf_t *, const char *);
static PhidgetReturnCode validInt(pconf_t *, pconf_t *, const char *);
static PhidgetReturnCode validDateTime(pconf_t *, pconf_t *, const char *);
static PhidgetReturnCode validKey(pconf_t *, pconf_t *, const char *);

static struct arg args[] = {
	{ "gen", "gen", "=", 1, validString},
	{ "startid", "id", ">=", 0, validInt},
	{ "endid", "id", "<=", 0, validInt},
	{ "startdate", "time", ">=", 2, validDateTime},
	{ "enddate", "time", "<=", 2, validDateTime},
	{ "key", "key", "=", 1, validKey},
	{ 0 }
};

static const char *dslog1_query = "select id, gen, time, key, val from dslog1";
//  "select id, gen, strftime('%Y-%m-%dT%H:%M:%f', time, 'localtime'), key, val from dslog1";

struct dslog1_row {
	int id;
	const uint8_t *gen;
	const uint8_t *time;
	const uint8_t *key;
	const uint8_t *val;
};

static PhidgetReturnCode
validKey(pconf_t *pc, pconf_t *dbpc, const char *key) {

	if (!Phidget_validDictionaryKey(key))
		return (EPHIDGET_INVALIDARG);

	if (!pconf_exists(dbpc, "dictionary.log.key.%s", key))
		return (EPHIDGET_NOENT);

	return (EPHIDGET_OK);
}

static PhidgetReturnCode
validDateTime(pconf_t *pc, pconf_t *dbpc, const char *val) {
	PhidgetReturnCode res;
	mostimestamp_t ts;

	res = mostimestamp_fromstring(MOS_IOP_IGNORE, val, &ts);
	if (res != 0)
		return (EPHIDGET_INVALIDARG);
	return (EPHIDGET_OK);
}

static PhidgetReturnCode
validInt(pconf_t *pc, pconf_t *dbpc, const char *val) {
	PhidgetReturnCode res;
	uint64_t u64;

	res = mos_strtou64(val, 0, &u64);
	if (res != 0)
		return (EPHIDGET_INVALIDARG);
	return (EPHIDGET_OK);
}

static PhidgetReturnCode
validString(pconf_t *pc, pconf_t *dbpc, const char *str) {

	if (!Phidget_validDictionaryKey(str))
		return (EPHIDGET_INVALIDARG);
	return (EPHIDGET_OK);
}

static int
getDictFormat(WebConnHandle wc) {
	const char *fmt;

	fmt = kvgetstrc(wc->query, "format", "JSON");
	if (mos_strcasecmp(fmt, "JSON") == 0)
		return (DICTFMT_JSON);
	if (mos_strcasecmp(fmt, "CSV") == 0)
		return (DICTFMT_CSV);

	return (DICTFMT_JSON); /* default, even with an invalid arg */
}

static PhidgetReturnCode
getDB(mosiop_t iop, pconf_t *pc, WebConnHandle wc, const char *dict, sqlite3 **db) {
	char dbpath[MOS_PATH_MAX];
	int err;

	mos_snprintf(dbpath, sizeof (dbpath), "%s/%s.db",
	  pconf_getstr(pc, DICTDBDIR, "phidget.dictionarystore.database.directory"), dict);

	err = sqlite3_open_v2(dbpath, db, SQLITE_OPEN_READONLY, NULL);
	if (err != SQLITE_OK) {
		wslogerr("failed to open database (%s): %d", dbpath, err);
		return (EPHIDGET_NOENT);
	}

	return (EPHIDGET_OK);
}

static PhidgetReturnCode
returnDB(sqlite3 *db) {
	int err;

	err = sqlite3_close_v2(db);
	if (err != SQLITE_OK)
		wslogerr("failed to close database connection: %d", err);
	return (EPHIDGET_OK);
}

#define QLEN	(sizeof (query) - (q - query))
#define ADDWHERE	do {						\
	if (where == 0) {							\
		mos_snprintf(q, QLEN, " where ");		\
		q += 7;									\
		where = 1;								\
	} else {									\
		mos_snprintf(q, QLEN, " and ");			\
		q += 5;									\
	}											\
} while (0)

#define ADDW(arg, val)	do {																		\
	switch ((arg)->type) {																			\
	case 0:																							\
		n = mos_snprintf(q, QLEN, " %s %s %s", (arg)->col, (arg)->cmp, (val));						\
		break;																						\
	case 1:																							\
		n = mos_snprintf(q, QLEN, " %s %s '%s'", (arg)->col, (arg)->cmp, (val));					\
		break;																						\
	case 2:																							\
		if (mos_endswith((val), "Z"))																\
			n = mos_snprintf(q, QLEN, " strftime('%%s', %s) %s strftime('%%s', '%s')",				\
			  (arg)->col, (arg)->cmp, (val));														\
		else																						\
			n = mos_snprintf(q, QLEN, " strftime('%%s', %s, 'localtime') %s strftime('%%s', '%s')",	\
			  (arg)->col, (arg)->cmp, (val));														\
		break;																						\
	}																								\
	if (n >= QLEN)																					\
		return (MOS_ERROR(iop, EPHIDGET_NOSPC, "query too large"));									\
	q += n;																							\
} while (0)

static PhidgetReturnCode
createDataStatement(mosiop_t iop, pconf_t *pc, pconf_t *dbpc, WebConnHandle wc, sqlite3 *db,
  sqlite3_stmt **stmt) {
	PhidgetReturnCode res;
	const char *aval;
	struct arg *arg;
	char query[512];
	int where;
	size_t n;
	char *q;
	int err;

	n = mos_strlcpy(query, dslog1_query, sizeof (query));
	q = query + n;

	where = 0;

	for (arg = args; arg->name != NULL; arg++) {
		aval = kvgetstrc(wc->query, arg->name, NULL);
		if (aval != NULL) {
			res = arg->check(pc, dbpc, aval);
			if (res != EPHIDGET_OK) {
				wslogwarn("invalid parameter '%s'='%s'", arg->name, aval);
				return (res);
			}
			ADDWHERE;
			ADDW(arg, aval);
		}
	}

	mos_strlcpy(q, " order by id;", QLEN);

	err = sqlite3_prepare_v2(db, query, -1, stmt, NULL);
	if (*stmt == NULL || err != SQLITE_OK)
		return (MOS_ERROR(iop, EPHIDGET_UNEXPECTED, "failed to create stmt from '%s':(%d)%s", query, err,
		  sqlite3_errstr(err)));

	return (EPHIDGET_OK);
}

/*
 * If an error occurs, log it, send it back to the user and return success.  Unless the error
 * cannot be related to the user providing invalid data or closing the connection, it is not
 * a server error.
 */
static PhidgetReturnCode
handleDictionaryAPIDataRequest(mosiop_t iop, pconf_t *pc, WebConnHandle wc, sqlite3 *db, pconf_t *dbpc,
  int fmt, const char *dict) {
	struct dslog1_row row;
	PhidgetReturnCode res;
	sqlite3_stmt *stmt;
	mostimestamp_t ts;
	char buf[1024];
	time_t current;
	int interval;
	time_t last;
	size_t n;
	int err;
	int cnt;

	stmt = NULL;

	res = createDataStatement(iop, pc, dbpc, wc, db, &stmt);
	if (res != EPHIDGET_OK) {
		wslogerr("createDataStatement() failed: %N", iop);
		wserror(iop, wc, 500, "Database Error", res, "failed to create database query");
		return (EPHIDGET_OK);
	}

	if (write_header(MOS_IOP_IGNORE, wc, fmt == DICTFMT_JSON ? "dict.json" : "dict.csv") != EPHIDGET_OK)
		goto done;

	n = mos_snprintf(buf, sizeof (buf), "{\"dictionary\":\"%s\",\"version\":1,\"data\":[",
	  kvgetstrc(wc->query, "name", ""));
	res = netConnWrite(MOS_IOP_IGNORE, wc->conn, buf, n);
	if (res != EPHIDGET_OK) {
		res = EPHIDGET_OK;
		goto done;
	}

	interval = kvgeti32(wc->query, "interval", -1);
	last = 0;

	cnt = 0;
	for (;;) {
		err = sqlite3_step(stmt);
		if (err != SQLITE_DONE && err != SQLITE_ROW) {
			MOS_ERROR(iop, EPHIDGET_UNEXPECTED, "failed to step query");
			goto done;
		}
		if (err == SQLITE_DONE)
			break;

		row.id = sqlite3_column_int(stmt, 0);
		row.gen = sqlite3_column_text(stmt, 1);
		row.time = sqlite3_column_text(stmt, 2);
		row.key = sqlite3_column_text(stmt, 3);
		row.val = sqlite3_column_text(stmt, 4);

		if (interval > 0) {
			if (mostimestamp_fromstring(MOS_IOP_IGNORE, (const char *)row.time, &ts) != 0) {
				wslogerr("failed to parse timestamp: %s", row.time);
				continue;
			}
			if (mostimestamp_toepoch(MOS_IOP_IGNORE, &ts, &current) != 0) {
				wslogerr("failed to calculate epoch from timestamp");
				continue;
			}
			if (current < last)
				continue;

			last = current + interval;
		}

		n = mos_snprintf(buf, sizeof (buf),
		  "%s{\"id\":%d,\"gen\":\"%s\",\"time\":\"%s\",\"key\":\"%s\",\"val\":\"%s\"}",
		  cnt > 0 ? "," : "", row.id, row.gen, row.time, row.key, row.val);

		/*
		 * Ignore rows that are too large (and as a result incomplete).
		 */
		if (n >= sizeof (buf))
			continue;

		res = netConnWrite(MOS_IOP_IGNORE, wc->conn, buf, n);
		if (res != EPHIDGET_OK) {
			res = EPHIDGET_OK;
			goto done;
		}
		cnt++;
	}
	n = mos_snprintf(buf, sizeof (buf), "],\"records\":%d}\n", cnt);
	netConnWrite(MOS_IOP_IGNORE, wc->conn, buf, n);

done:

	if (stmt)
		sqlite3_finalize(stmt);

	if (res != EPHIDGET_OK)
		wserror(iop, wc, 500, "Internal Error", res, "Failed to execute query");
	return (res);
}

static PhidgetReturnCode
handleDictionaryAPIGetRequest(mosiop_t iop, pconf_t *pc, WebConnHandle wc, const char *dict) {
	PhidgetReturnCode res;
	char buf[16384];

	res = pconf_renderjson(pc, buf, sizeof (buf));
	if (res != EPHIDGET_OK)
		return (MOS_ERROR(iop, res, "failed to render dictionary json: %s", dict));

	res = write_header(iop, wc, "dict.json");
	if (res != EPHIDGET_OK)
		return (MOS_ERROR(iop, res, "failed to write HTTP header"));

	res = netConnWrite(iop, wc->conn, buf, mos_strlen(buf));
	if (res != EPHIDGET_OK)
		return (MOS_ERROR(iop, res, "failed to write dictionary data"));

	return (EPHIDGET_OK);
}

static PhidgetReturnCode
readDictionaryCfg(pconf_t **dbpc, const char *fmt, ...) {
	char path[MOS_PATH_MAX];
	PhidgetReturnCode res;
	char dbbuf[65536];
	char errbuf[128];
	size_t dbbufsz;
	va_list va;
	size_t n;
	int err;
	int i;

	va_start(va, fmt);
	n = mos_vsnprintf(path, sizeof (path), fmt, va);
	va_end(va);
	if (n >= sizeof (path))
		return (EPHIDGET_NOSPC);

	for (i = 0; i <= 3; i++) {
		dbbufsz = sizeof (dbbuf);
		err = mos_file_readx(MOS_IOP_IGNORE, dbbuf, &dbbufsz, "%s", path);
		if (err == 0)
			break;

		if (err == MOSN_BUSY && i < 3) {
			wslogdebug("dictionary '%s' busy while reading: trying again", path);
			mos_usleep(250000);
			continue;
		}
		wslogerr("failed to read dictionary '%s': %d", path, err);
		return (err);
	}

	dbbuf[dbbufsz] = '\0';
	res = pconf_parsepcs(dbpc, errbuf, sizeof (errbuf), dbbuf, dbbufsz);
	if (res != EPHIDGET_OK) {
		wslogerr("failed to parse database config '%s': %s\n", path, errbuf);
		wslogdebug("failed to parse [%s]", dbbuf);
	}

	return (res);
}

static PhidgetReturnCode
handleDictionaryAPIRequest(mosiop_t iop, pconf_t *pc, WebConnHandle wc, int *keepalive) {
	PhidgetReturnCode res;
	const char *action;
	const char *file;
	const char *dict;
	pconf_t *dbpc;
	sqlite3 *db;
	int fmt;

	fmt = getDictFormat(wc);
	*keepalive = 0;
	db = NULL;

	action = kvgetstrc(wc->query, "action", NULL);
	if (action == NULL)
		action = "data";

	dict = kvgetstrc(wc->query, "name", NULL);
	if (dict == NULL) {
		wserror(iop, wc, 422, "Missing Parameter", EPHIDGET_INVALIDARG, "%s", "missing name");
		return (EPHIDGET_OK);
	}

	file = pconf_getstr(pc, DICTCONFIGDIR, "phidget.feature.dictionary.directory");
	res = readDictionaryCfg(&dbpc, "%s/%s.dpc", file, dict);
	if (res != EPHIDGET_OK) {
		wserror(iop, wc, 500, "Missing Database", EPHIDGET_NOENT, "missing database config %s", dict);
		return (EPHIDGET_OK);
	}

	if (action == NULL || mos_strcmp(action, "data") == 0) {
		res = getDB(iop, pc, wc, dict, &db);
		if (res != EPHIDGET_OK) {
			if (res == EPHIDGET_NOENT) {	/* XXX Return Empty dataset */
				if (write_header(MOS_IOP_IGNORE, wc, wc->uri) == EPHIDGET_OK) {
					if (fmt == DICTFMT_JSON) {
						netConnWrite(MOS_IOP_IGNORE, wc->conn, "{}", 2);
					}
				}
				res = EPHIDGET_OK;
				goto done;
			}
			wserror(iop, wc, 500, "Database Error", res, "failed to access database");
			res = EPHIDGET_OK;
			goto done;
		}

		res = handleDictionaryAPIDataRequest(iop, pc, wc, db, dbpc, fmt, dict);
	} else if (mos_strcmp(action, "get") == 0) {
		res = handleDictionaryAPIGetRequest(iop, dbpc, wc, dict);
	} else {
		wserror(iop, wc, 400, "Invalid Parameter", EPHIDGET_INVALIDARG, "invalid action '%s'", action);
	}

done:

	pconf_release(&dbpc);
	if (db)
		returnDB(db);
	return (res);
}

PhidgetReturnCode
handleAPIRequest(mosiop_t iop, pconf_t *pc, WebConnHandle wc, int *keepalive) {
	const char *basename;

	basename = mos_basename(wc->uri);
	if (mos_strncmp(basename, DICTIONARYAPI, mos_strlen(DICTIONARYAPI)) != 0) {
		write_noentheader(iop, wc, wc->uri);
		return (0);
	}

	return (handleDictionaryAPIRequest(iop, pc, wc, keepalive));
}